#!/usr/bin/env python3

import multiprocessing
import os
import random
import time
from glob import glob
from shutil import rmtree
from subprocess import getstatusoutput as cmd
from sys import argv


def readInstances(file):
    inst = []
    f = open(file, 'r')
    for l in f:
        if len(l) < 5 or l[0] == '#':
            continue
        i = [e.strip() for e in l.split('|')][1:-1]
        i[2] = '../' + i[2]
        i[3] = '../' + i[3]
        inst.append(tuple(i))
    return inst


def generate(i):
    name = i[0] + '_' + '_'.join(i[3].split('/')[-3:]) + '_' + i[4]
    dir = 'TMP_' + name
    cnfFile = outDir + '/' + i[1] + '_' + name + '.cnf'
    os.mkdir(dir)
    c = 'cd ' + dir + ' && '

    if 'M' in i[0]:
        c += '../Mp -A 1 -S 1 -O -F ' + i[4] + \
            ' -T ' + i[4] + ' ' + i[2] + " " + i[3]
        if 'MS' in i[0]:
            c += ' -P 0'

    if 'P' in i[0]:
        c += '../fd/src/translate/translate.py ' + i[2] + ' ' + i[3] + ' && '
        c += '../pasar_glucose -v 9 -s 4 -mi 1 -i 0 -ml -1 -im ' + \
            i[4] + " -of ../" + cnfFile + " output.sas"

    if 'H' in i[0]:
        c += '../treerexx -of -d=' + i[4] + \
            ' -D=' + i[4] + ' ' + i[2] + ' ' + i[3]

    print('start ' + name)
    t1 = time.time()
    log = c
    r = cmd(c)
    log += '\n' + r[1]
    t = time.time() - t1
    log += '\ngenerationTime: ' + str(t)

    # tree-rexx returns non zero for unsat makespans
    if r[0] != 0 and ('H' not in i[0] or 'UNSAT' not in i[1]):
        return 'Error: could not generate cnfFile see ' + logDir + '/' + name + '.log'

    if 'M' in i[0]:
        for l in log.split('\n'):
            if "Writing" in l:
                resFile = l.split(' ')[1]
        os.rename(dir + '/' + resFile, cnfFile)
    if 'H' in i[0]:
        os.rename(dir + '/f.cnf', cnfFile)

    rmtree(dir)
    with open(logDir + '/' + name + '.log', 'w') as f:
        f.write(log)
    return 'done  ' + name


def runInParallel(commands, threads):
    with multiprocessing.Pool(threads) as p:
        for i in p.imap_unordered(generate, commands):
            print(i)


cmd('./build.sh')
logDir = 'log'
outDir = 'planning'

if not os.path.exists(logDir):
    os.makedirs(logDir)
if not os.path.exists(outDir):
    os.makedirs(outDir)

# clean if previous was aborted
for tmp in glob('TMP_*'):
    rmtree(tmp)

instances = readInstances('instances.org')


# be aware of memory problems
n = 1
if len(argv) > 1:
    n = int(argv[1])

    # advanced load balancing
    random.seed(42)
    random.shuffle(instances)

runInParallel(instances, n)
