
#ifndef DOMPASCH_TREE_REXX_HTN_OP_H
#define DOMPASCH_TREE_REXX_HTN_OP_H

#include <vector>

#include "signature.h"


class HtnOp {

protected:
    int _id;
    std::vector<int> _args;

    SigSet _preconditions;
    SigSet _effects;

public:
    HtnOp() {}
    HtnOp(int id, std::vector<int> args) : _id(id), _args(args) {}
    HtnOp(const HtnOp& op) : _id(op._id), _args(op._args), _preconditions(op._preconditions), _effects(op._effects) {}

    void setPreconditions(const SigSet& set) {
        _preconditions = set;
    }
    void addPrecondition(const Signature& sig) {
        _preconditions.insert(sig);
    }
    void addEffect(const Signature& sig) {
        _effects.insert(sig);
    }
    void addArgument(int arg) {
        _args.push_back(arg);
    }
    void removeInconsistentEffects() {
        // Collect all neg. preconds for which the pos. precond is contained, too
        SigSet inconsistentEffs;
        for (Signature sig : _effects) {
            if (sig._negated && _effects.count(sig.abs())) {
                inconsistentEffs.insert(sig);
            }
        }
        // Remove each such neg. precond
        for (Signature sig : inconsistentEffs) {
            _effects.erase(sig);
        }
    }

    virtual HtnOp substitute(const std::unordered_map<int, int>& s) const {
        HtnOp op;
        op._id = _id;
        op._args.resize(_args.size());
        for (int i = 0; i < _args.size(); i++) {
            if (s.count(_args[i])) op._args[i] = s.at(_args[i]);
            else op._args[i] = _args[i];
        }
        for (Signature sig : _preconditions) {
            Signature sigSubst = sig.substitute(s);
            op.addPrecondition(sigSubst);
        }
        for (Signature sig : _effects) {
            Signature sigSubst = sig.substitute(s);
            op.addEffect(sigSubst);
        }
        return op;
    }

    const SigSet& getPreconditions() const {
        return _preconditions;
    }
    const SigSet& getEffects() const {
        return _effects;
    }
    const std::vector<int>& getArguments() const {
        return _args;
    }
    Signature getSignature() const {
        return Signature(_id, _args);
    }

    HtnOp& operator=(const HtnOp& op) {
        _id = op._id;
        _args = op._args;
        _preconditions = op._preconditions;
        _effects = op._effects;
        return *this;
    }
};


#endif